@@ -1,12 +1,11 @@
-/*************************************************************************
+/*
+ * vmur - Work with z/VM spool file queues (reader, punch, printer)
  *
- * File...........: s390-tools/vmur/vmur.c
- * Author(s)......: Frank Munzert <munzert@de.ibm.com>
- *                  Michael Holzheu <holzheu@de.ibm.com>
+ * Copyright IBM Corp. 2007, 2017
  *
- * Copyright IBM Corp. 2007
- *
- *************************************************************************/
+ * s390-tools is free software; you can redistribute it and/or modify
+ * it under the terms of the MIT license. See LICENSE for details.
+ */
 
 #include <ctype.h>
 #include <errno.h>
@@ -23,27 +22,31 @@
 #include <iconv.h>
 #include <sys/types.h>
 #include <dirent.h>
+#include <sys/sysmacros.h>
 #include <sys/ioctl.h>
-#include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/file.h>
 #include <ctype.h>
 #include <linux/types.h>
-#include "zt_common.h"
+
+#include "lib/vmdump.h"
+#include "lib/zt_common.h"
+#include "lib/util_libc.h"
+#include "lib/vmcp.h"
+
 #include "vmur.h"
 
+#define	CP_PREFIX_LEN	11
+
 /* Program name */
 static char *prog_name;
 
 /* Short description */
 static const char tool_name[] =
-	"vmur: Control virtual reader, punch and printer";
+	"vmur: Control virtual reader, punch, and printer";
 
 /* Copyright notice */
-static const char copyright_notice[] = "Copyright IBM Corp. 2007";
-
-/* Structure for all global settings */
-struct vmur vmur_info;
+static const char copyright_notice[] = "Copyright IBM Corp. 2007, 2017";
 
 /*
  * Structure vmur is used to store the command line parameters and
@@ -81,15 +84,27 @@ struct vmur {
 	enum spoolfile_fmt spoolfile_fmt;
 	struct sigaction sigact;
 	iconv_t iconv;
-};
+	int   lock_fd;
+	/* ur device spool state */
+	char  spool_restore_cmd[MAXCMDLEN];
+	int   spool_restore_needed;
+	char  spool_class;
+	int   spool_class_specified;
+	char  spool_dest[9];
+	int   spool_dest_specified;
+	char  spool_form[9];
+	int   spool_form_specified;
+	char  spool_dist[9];
+	int   spool_dist_specified;
+} vmur_info;
 
 /*
  * Print version information.
  */
 static void print_version (void)
 {
-	printf("%s version %s.\n", tool_name, RELEASE_STRING);
-	printf("%s.\n", copyright_notice);
+	printf("%s version %s\n", tool_name, RELEASE_STRING);
+	printf("%s\n", copyright_notice);
 }
 
 /*
@@ -147,7 +162,7 @@ static char HELP_TEXT[] =
 "       vmur order   [OPTIONS] [SPOOLID]\n"
 "       vmur list    [OPTIONS] [SPOOLID]\n"
 "\n"
-"Control virtual reader, punch and printer. Available commands are:\n"
+"Control virtual reader, punch, and printer. Available commands are:\n"
 "  * REceive: Receive spool files from reader queue\n"
 "  * PUNch:   Punch a file to punch queue\n"
 "  * PRint:   Print a file to printer queue\n"
@@ -170,6 +185,7 @@ static char HELP_TEXT[] =
 "-O, --stdout             Write spool file to stdout.\n"
 "-f, --force              Overwrite files without prompt.\n"
 "-H, --hold               Hold spool file in reader after receive.\n"
+"-C, --class              Specify the spool class to match a reader file.\n"
 "\n"
 "Options for 'punch' and 'print' command:\n"
 "\n"
@@ -186,6 +202,12 @@ static char HELP_TEXT[] =
 "-n, --node               Remote node to send the file to.\n"
 "                         If omitted: Your local VM node.\n"
 "-N, --name               Name of new spool file.\n"
+"-f, --force              Convert file name to valid spool file name\n"
+"                         automatically without prompt.\n"
+"-C, --class              Spool class to be assigned to the created spool file.\n"
+"    --form               Form to be assigned to the created spool file.\n"
+"    --dest               Destination to be assigned to the created spool file.\n"
+"    --dist               Distribution code for the resulting spool file.\n"
 "\n"
 "Options for 'purge' command:\n"
 "\n"
@@ -198,7 +220,7 @@ static char HELP_TEXT[] =
 
 static void usage(void)
 {
-	printf(HELP_TEXT);
+	printf("%s", HELP_TEXT);
 }
 
 /*
@@ -222,30 +244,12 @@ fail:
 	ERR_EXIT("Could not initialize signal handler (errno = %i)\n", errno);
 }
 
-/*
- * Read at most COUNT bytes from FD into memory at location BUF.
- * Return number of bytes read on success, -1 on error.
- */
-static ssize_t read_buf(int fd, char *buf, size_t count)
-{
-	ssize_t rc, len;
-
-	for (len = 0; len < count; len += rc) {
-		rc = read(fd, &buf[len], count - len);
-		if (rc == -1)
-			return -1;
-		if (rc == 0)
-			break;
-	}
-	return len;
-}
-
 /*
  * Strip leading CP header from error message
  */
 static void strip_cperr(char *buf)
 {
-	int i;
+	unsigned int i;
 
 	for (i = 0; i < strlen(buf) - CP_PREFIX_LEN; i++) {
 		if (strncmp(buf + i, "HCP", 3) == 0) {
@@ -258,68 +262,78 @@ static void strip_cperr(char *buf)
 	}
 }
 
-int _cpcmd(char *cpcmd, char **resp, int *rc, int retry, int upper)
+/*
+ * Handle CP error message and exit
+ */
+static void cperr_exit(char *cpcmd, int cprc, char *buf)
 {
-	int fd, len, cprc, bufsize = VMCP_BUFSIZE;
-	char *buf;
+	if (strlen(buf) <= CP_PREFIX_LEN)
+		ERR_EXIT("CP command '%s' failed with rc=%i\n", cpcmd, cprc);
+
+	strip_cperr(buf);
+	ERR_EXIT("%s\n", buf);
+}
+
+static void _cpcmd(char *cpcmd, char **resp, int *rc, int retry, int upper)
+{
+	struct vmcp_parm cp;
 	char cmd[MAXCMDLEN];
+	int ret;
 
 	strcpy(cmd, cpcmd);
-	if (upper)
-		to_upper(cmd);
+	cp.cpcmd = cmd;
+	cp.do_upper = upper;
+	cp.buffer_size = VMCP_DEFAULT_BUFSZ;
 
-	fd = open(VMCP_DEVICE_NODE, O_RDWR);
-	if (fd == -1)
-		ERR_EXIT("Could not issue cp command: \"%s\"\n"
+retry:
+	ret = vmcp(&cp);
+
+	switch (ret) {
+	case VMCP_ERR_OPEN:
+		ERR_EXIT("Could not issue CP command: \"%s\"\n"
 			 "Ensure that vmcp kernel module is loaded!\n", cmd);
 
-	do {
-		if (ioctl(fd, VMCP_SETBUF, &bufsize) == -1)
-			goto fail;
-
-		if (write(fd, cmd, strlen(cmd)) == -1)
-			goto fail;
-
-		if (ioctl(fd, VMCP_GETCODE, &cprc) == -1)
-			goto fail;
-
-		if (ioctl(fd, VMCP_GETSIZE, &len) == -1)
-			goto fail;
-
-		if (len <= bufsize)
-			break;
-		else if (retry)
-			bufsize = len;
-		else
-			ERR_EXIT("Not enough buffer space (%i/%i) for CP "
-				 "command '%s'.\nSorry, please issue command "
-				 "on your 3270 console!\n", len, bufsize, cmd);
-	} while (1);
-
-	buf = malloc(len + 1);
-	if (!buf)
-		ERR_EXIT("Out of memory for CP command '%s'\n", cmd);
-
-	memset(buf, 0, len + 1);
-	if (read_buf(fd, buf, len) == -1)
+	case VMCP_ERR_SETBUF:
 		goto fail;
 
+	case VMCP_ERR_WRITE:
+		goto fail;
+
+	case VMCP_ERR_GETCODE:
+		goto fail;
+
+	case VMCP_ERR_GETSIZE:
+		goto fail;
+
+	case VMCP_ERR_READ:
+		goto fail;
+
+	case VMCP_ERR_TOOSMALL:
+		if (retry) {
+			cp.buffer_size = cp.response_size;
+			free(cp.response);
+			goto retry;
+		}
+		ERR_EXIT("Not enough buffer space (%u/%u) for CP "
+			 "command '%s'.\nSorry, please issue command "
+			 " on your 3270 console!\n", cp.response_size,
+			 cp.buffer_size, cmd);
+	}
+
 	if (rc == NULL) {
-		if (cprc != 0) {
+		if (cp.cprc != 0) {
 			/* caller wants us to handle the error */
-			strip_cperr(buf);
-			ERR_EXIT("%s\n", buf);
+			cperr_exit(cmd, cp.cprc, cp.response);
 		}
 	} else {
-		*rc = cprc;
+		*rc = cp.cprc;
 	}
 
 	if (resp)
-		*resp = buf;
+		*resp = cp.response;
 	else
-		free(buf);
-	close(fd);
-	return strlen(buf);
+		free(cp.response);
+	return;
 
 fail:
 	ERR_EXIT("CP command '%s' failed.\n", cmd);
@@ -333,9 +347,9 @@ fail:
  * @retry: retry = 1 -> CP command can be retried.
  * This function converts the command string to uppercase.
  */
-int cpcmd(char *cpcmd, char **resp, int *rc, int retry)
+static void cpcmd(char *cpcmd, char **resp, int *rc, int retry)
 {
-	return _cpcmd(cpcmd, resp, rc, retry, 1);
+	_cpcmd(cpcmd, resp, rc, retry, 1);
 }
 
 /*
@@ -346,9 +360,9 @@ int cpcmd(char *cpcmd, char **resp, int *rc, int retry)
  * @retry: retry = 1 -> CP command can be retried.
  * This function does not convert the command string to uppercase.
  */
-int cpcmd_cs(char *cpcmd, char **resp, int *rc, int retry)
+static void cpcmd_cs(char *cpcmd, char **resp, int *rc, int retry)
 {
-	return _cpcmd(cpcmd, resp, rc, retry, 0);
+	_cpcmd(cpcmd, resp, rc, retry, 0);
 }
 
 /*
@@ -358,13 +372,13 @@ static int get_minor(char *path)
 {
 	FILE *fd;
 	char buf[20];
-	int major, minor;
+	int major, minor, rc;
 
 	fd = fopen(path, "r");
 	if (!fd)
 		ERR_EXIT("Could not open %s (err = %i)\n", path, errno);
-	fread(buf, sizeof(buf), 1, fd);
-	if (ferror(fd))
+	rc = fread(buf, sizeof(buf), 1, fd);
+	if (rc == -1)
 		ERR_EXIT("Could not read %s (err = %i)\n", path, errno);
 	fclose(fd);
 
@@ -407,10 +421,11 @@ static int get_ur_devno(int minor)
 		ERR_EXIT("Could not read %s (err = %i)\n",
 			    SYSFS_CLASS_DIR, errno);
 	else
-		ERR_EXIT("Device not online\n");
+		ERR_EXIT("Device is not online\n");
 found:
 	if (sscanf(bus_id, "0.0.%x", &devno) != 1)
-		ERR_EXIT("Could not extract devno from %s\n", direntp->d_name);
+		ERR_EXIT("Could not extract device number from %s\n",
+			 direntp->d_name);
 	closedir(fd);
 	return devno;
 }
@@ -434,7 +449,7 @@ static dev_t get_node_dev(char *devnode)
 /*
  * Extract major of vmur driver
  */
-static int get_driver_major(void)
+static int get_driver_major(unsigned long *major)
 {
 	FILE *fh;
 	char string[PROC_DEVICES_FILE_WIDTH];
@@ -449,7 +464,8 @@ static int get_driver_major(void)
 	while (fscanf(fh, "%s", string) != EOF) {
 		if (strcmp(string, "vmur") == 0) {
 			fclose(fh);
-			return atoi(last_string);
+			*major = atoi(last_string);
+			return 0;
 		} else {
 			strcpy(last_string, string);
 		}
@@ -458,17 +474,154 @@ static int get_driver_major(void)
 	ERR_EXIT("Unit record device driver not loaded.\n");
 }
 
+/*
+ * Copy printable characters
+ */
+static void strncpy_graph(char *dest, const char *src, size_t len)
+{
+	size_t n;
+
+	for (n = 0; n < len; n++) {
+		if (!isgraph(src[n]))
+			break;
+		dest[n] = src[n];
+	}
+	dest[n] = '\0';
+}
+
+/*
+ * Create CP command to restore spooling options
+ */
+static void save_spool_options(struct vmur *info)
+{
+	char cmd[MAXCMDLEN], *resp;
+	char cl, value[9], *tmp;
+	int n;
+
+	/* Retrieve spooling options for ur device */
+	sprintf(cmd, "QUERY VIRTUAL %X", info->devno);
+	cpcmd(cmd, &resp, NULL, 0);
+
+	/* Prepare CP spool restore command */
+	n = sprintf(info->spool_restore_cmd, "SPOOL %X", info->devno);
+
+	/* Save the CLASS value if required */
+	if (info->spool_class_specified) {
+		cl = resp[13];
+		n += sprintf(info->spool_restore_cmd + n, " CL %c", cl);
+	}
+
+	/* Save FORM value if required */
+	if (info->spool_form_specified) {
+		tmp = strstr(resp, "FORM ");
+		if (tmp == NULL)
+			ERR_EXIT("Could not retrieve value for the FORM "
+				 "spooling option\n");
+		strncpy_graph(value, tmp + 5, 8);
+		n += sprintf(info->spool_restore_cmd + n, " FORM %s", value);
+	}
+
+	/* Save DEST value if required */
+	if (info->spool_dest_specified) {
+		tmp = strstr(resp, "DEST ");
+		if (tmp == NULL)
+			ERR_EXIT("Could not retrieve value for the DEST "
+				 "spooling option\n");
+		strncpy_graph(value, tmp + 5, 8);
+		n += sprintf(info->spool_restore_cmd + n, " DEST %s", value);
+	}
+
+	/* Save DIST value if required */
+	if (info->spool_dist_specified) {
+		tmp = strstr(resp, "DIST ");
+		if (tmp == NULL)
+			ERR_EXIT("Could not retrieve value for the DIST "
+				 "spooling option\n");
+		strncpy_graph(value, tmp + 5, 8);
+		n += sprintf(info->spool_restore_cmd + n, " DIST %s", value);
+	}
+
+	free(resp);
+	++info->spool_restore_needed;
+}
+
+/*
+ * Restore saved spooling options for a ur device
+ */
+static void restore_spool_options(struct vmur *info)
+{
+	if (!info->spool_restore_needed)
+		return;
+
+	cpcmd(info->spool_restore_cmd, NULL, NULL, 0);
+	--info->spool_restore_needed;
+}
+
+/*
+ * Returns non-zero if spool options have to be saved, changed, and restored,
+ * otherwise zero is returned.
+ */
+static int require_spool_setup(struct vmur *info)
+{
+	return !!(info->spool_class_specified ||
+		  info->spool_form_specified ||
+		  info->spool_dest_specified ||
+		  info->spool_dist_specified);
+}
+
+/*
+ * Setup spooling options for a ur device
+ */
+static void setup_spool_options(struct vmur *info)
+{
+	char cmd[MAXCMDLEN];
+	int n, rc;
+
+	/*
+	 * Check if spool options must be changed.  If so, save the current
+	 * spool option values and restore them at program exit.
+	 */
+	if (!require_spool_setup(info))
+		return;
+
+	/* Save spool options */
+	save_spool_options(info);
+
+	/* Change spool options */
+	n = sprintf(cmd, "SPOOL %X", info->devno);
+	if (info->spool_class_specified)
+		n += sprintf(cmd + n, " CLASS %c", info->spool_class);
+	if (info->spool_form_specified)
+		n += sprintf(cmd + n, " FORM %s", info->spool_form);
+	if (info->spool_dest_specified)
+		n += sprintf(cmd + n, " DEST %s", info->spool_dest);
+	if (info->spool_dist_specified)
+		n += sprintf(cmd + n, " DIST %s", info->spool_dist);
+
+	cpcmd(cmd, NULL, &rc, 0);
+	if (rc)
+		ERR_EXIT("Could not set spooling options (rc=%i)\n", rc);
+}
+
 /*
  * Setup and check ur device
  */
 static void setup_ur_device(struct vmur *info)
 {
-	int driver_major;
+	unsigned long driver_major;
 	dev_t node_dev;
+	int rc;
+
+	/*
+	 * Few vmur commands do not require a particular device node and,
+	 * therefore, no device setup is required.
+	 */
+	if (!strlen(info->devnode))
+		return;
 
-	driver_major = get_driver_major();
 	node_dev = get_node_dev(info->devnode);
-	if ((driver_major != -1) && (driver_major != major(node_dev)))
+	rc = get_driver_major(&driver_major);
+	if ((rc == 0) && (driver_major != major(node_dev)))
 		ERR_EXIT("'%s' is not a unit record device.\n",
 			 info->devnode);
 	info->devno = get_ur_devno(minor(node_dev));
@@ -481,6 +634,7 @@ static void init_info(struct vmur *info)
 {
 	memset(info, 0, sizeof(struct vmur));
 	strcpy(info->queue, "rdr");
+	info->lock_fd = -1;
 }
 
 /*
@@ -551,6 +705,9 @@ static void set_queue(struct vmur *info, char *queue)
 	return;
 }
 
+/*
+ * Set block mode and the separator and padding byte
+ */
 static void set_blocked(struct vmur *info, char *blocked)
 {
 	if (sscanf(blocked, "0x%x,0x%x", &info->blocked_separator,
@@ -567,6 +724,22 @@ fail:
 		 "'0xSS,0xPP'.\n");
 }
 
+/*
+ * Set spool class value
+ */
+static void set_spool_class(struct vmur *info, const char *val, int is_rdr)
+{
+	char cl = val[0];
+
+	if (strlen(val) > 1 || (!isalnum(cl) && cl != '*'))
+		ERR_EXIT("Class must be one of A through Z, 0 through 9, "
+			 "or an asterisk (*)\n");
+	if (cl == '*' && !is_rdr)
+		ERR_EXIT("The asterisk (*) class is only valid for readers\n");
+	info->spool_class = toupper(cl);
+	++info->spool_class_specified;
+}
+
 /*
  * Parse the command line: General options
  */
@@ -601,11 +774,13 @@ static void parse_opts_receive(struct vmur *info, int argc, char *argv[])
 		{ "stdout",      no_argument,       NULL, 'O'},
 		{ "force",       no_argument,       NULL, 'f'},
 		{ "hold",        no_argument,       NULL, 'H'},
+		{ "convert",     no_argument,       NULL, 'c'},
 		{ "device",      required_argument, NULL, 'd'},
 		{ "blocked",     required_argument, NULL, 'b'},
+		{ "class",       required_argument, NULL, 'C'},
 		{ 0,             0,                 0,    0  }
 	};
-	static const char option_string[] = "vhtOfHd:b:";
+	static const char option_string[] = "vhtOfHcd:b:C:";
 
 	strcpy(info->devnode, VMRDR_DEVICE_NODE);
 	while (1) {
@@ -635,6 +810,15 @@ static void parse_opts_receive(struct vmur *info, int argc, char *argv[])
 		case 'H':
 			++info->hold_specified;
 			break;
+		case 'c':
+			fprintf(stderr, "%s: Option '-c' is no longer valid. "
+				"Use zgetdump for dump file conversion.\n",
+				prog_name);
+			std_usage_exit();
+			break;
+		case 'C':
+			set_spool_class(info, optarg, 1);
+			break;
 		default:
 			std_usage_exit();
 		}
@@ -647,12 +831,18 @@ static void parse_opts_receive(struct vmur *info, int argc, char *argv[])
 	CHECK_SPEC_MAX(info->devnode_specified, 1, "devnode");
 	CHECK_SPEC_MAX(info->file_name_specified, 1, "file");
 	CHECK_SPEC_MAX(info->blocked_specified, 1, "blocked");
+	CHECK_SPEC_MAX(info->force_specified, 1, "force");
+	CHECK_SPEC_MAX(info->hold_specified, 1, "hold");
+	CHECK_SPEC_MAX(info->stdout_specified, 1, "stdout");
+	CHECK_SPEC_MAX(info->spool_class_specified, 1, "class");
 
 	if (info->stdout_specified && info->file_name_specified)
-		ERR_EXIT("File name not allowed, when --stdout specifed!\n");
-	if (info->blocked_specified && info->text_specified)
-		ERR_EXIT("Conflicting options: -b together with -t "
-			 "specified\n");
+		ERR_EXIT("File name not allowed, when --stdout specified!\n");
+	if (info->blocked_specified + info->text_specified)
+		ERR_EXIT("Conflicting options: -b and -t are mutually "
+			 "exclusive.\n");
+	if (!info->spool_class_specified)
+		set_spool_class(info, "*", 1);
 }
 
 /*
@@ -689,7 +879,7 @@ static int invalid_operand(char *operand)
 /*
  * Check whether user is in the CP directory
  */
-static int check_local_user(char *user)
+static int check_local_user(const char *user)
 {
 	char cmd[MAXCMDLEN];
 	int cprc;
@@ -728,7 +918,7 @@ static void set_spoolfile_name(struct vmur *info, char *name)
 				break;
 		}
 	}
-	if ((i > 0) && (i < strlen(spoolfile_name) - 1)) {
+	if ((i > 0) && (i < (int) strlen(spoolfile_name) - 1)) {
 		strcpy(spoolfile_type, spoolfile_name + i + 1);
 		spoolfile_name[i] = 0;
 		++info->spoolfile_type_specified;
@@ -774,9 +964,13 @@ static void parse_opts_punch_print(struct vmur *info, int argc, char *argv[])
 		{ "device",      required_argument, NULL, 's'},
 		{ "blocked",     required_argument, NULL, 'b'},
 		{ "name",        required_argument, NULL, 'N'},
+		{ "class",       required_argument, NULL, 'C'},
+		{ "dest",        required_argument, NULL, 'D'},
+		{ "form",        required_argument, NULL, 'F'},
+		{ "dist",	 required_argument, NULL, 'I'},
 		{ 0,             0,                 0,    0  }
 	};
-	static const char option_string[] = "vhtrfu:n:d:b:N:";
+	static const char option_string[] = "vhtrfu:n:d:b:N:C:";
 
 	if (info->action == PUNCH) {
 		strcpy(info->devnode, VMPUN_DEVICE_NODE);
@@ -827,6 +1021,30 @@ static void parse_opts_punch_print(struct vmur *info, int argc, char *argv[])
 			else
 				strcpy(info->node, optarg);
 			break;
+		case 'C':
+			set_spool_class(info, optarg, 0);
+			break;
+		case 'D':
+			++info->spool_dest_specified;
+			if (invalid_operand(optarg))
+				ERR_EXIT("Invalid destination: %s\n", optarg);
+			else
+				strcpy(info->spool_dest, optarg);
+			break;
+		case 'F':
+			++info->spool_form_specified;
+			if (invalid_operand(optarg))
+				ERR_EXIT("Invalid form: %s\n", optarg);
+			else
+				strcpy(info->spool_form, optarg);
+			break;
+		case 'I':
+			++info->spool_dist_specified;
+			if (invalid_operand(optarg))
+				ERR_EXIT("Invalid distribution code: %s\n", optarg);
+			else
+				strcpy(info->spool_dist, optarg);
+			break;
 		default:
 			std_usage_exit();
 		}
@@ -840,6 +1058,10 @@ static void parse_opts_punch_print(struct vmur *info, int argc, char *argv[])
 	CHECK_SPEC_MAX(info->user_specified, 1, "user");
 	CHECK_SPEC_MAX(info->node_specified, 1, "node");
 	CHECK_SPEC_MAX(info->blocked_specified, 1, "blocked");
+	CHECK_SPEC_MAX(info->spool_class_specified, 1, "class");
+	CHECK_SPEC_MAX(info->spool_form_specified, 1, "form");
+	CHECK_SPEC_MAX(info->spool_dest_specified, 1, "dest");
+	CHECK_SPEC_MAX(info->spool_dist_specified, 1, "dist");
 
 	if (info->user_specified && !info->rdr_specified)
 		ERR_EXIT("--user without --rdr specified\n");
@@ -875,9 +1097,12 @@ static void parse_opts_purge(struct vmur *info, int argc, char *argv[])
 		{ "help",        no_argument,       NULL, 'h'},
 		{ "force",       no_argument,       NULL, 'f'},
 		{ "queue",       no_argument,       NULL, 'q'},
+		{ "class",       required_argument, NULL, 'C'},
+		{ "dest",        required_argument, NULL, 'D'},
+		{ "form",        required_argument, NULL, 'F'},
 		{ 0,             0,                 0,    0  }
 	};
-	static const char option_string[] = "fvhq:";
+	static const char option_string[] = "fvhq:C:";
 
 	while (1) {
 		opt = getopt_long(argc, argv, option_string,
@@ -893,12 +1118,32 @@ static void parse_opts_purge(struct vmur *info, int argc, char *argv[])
 			++info->queue_specified;
 			set_queue(info, optarg);
 			break;
+		case 'C':
+			set_spool_class(info, optarg, 0);
+			break;
+		case 'D':
+			++info->spool_dest_specified;
+			if (invalid_operand(optarg))
+				ERR_EXIT("Invalid destination: %s\n", optarg);
+			else
+				strcpy(info->spool_dest, optarg);
+			break;
+		case 'F':
+			++info->spool_form_specified;
+			if (invalid_operand(optarg))
+				ERR_EXIT("Invalid form: %s\n", optarg);
+			else
+				strcpy(info->spool_form, optarg);
+			break;
 		default:
 			std_usage_exit();
 		}
 	}
 	CHECK_SPEC_MAX(info->force_specified, 1, "force");
 	CHECK_SPEC_MAX(info->queue_specified, 1, "queue");
+	CHECK_SPEC_MAX(info->spool_class_specified, 1, "class");
+	CHECK_SPEC_MAX(info->spool_form_specified, 1, "form");
+	CHECK_SPEC_MAX(info->spool_dest_specified, 1, "dest");
 	set_spoolid(info, argv, argc, optind + 1, 0);
 }
 
@@ -971,7 +1216,7 @@ static void parse_opts_list(struct vmur *info, int argc, char *argv[])
 /*
  * Parse the command line: Default options
  */
-static void parse_opts_default(struct vmur *info, int argc, char *argv[])
+static void parse_opts_default(int argc, char *argv[])
 {
 	int opt, index;
 	static struct option long_options[] = {
@@ -1036,7 +1281,7 @@ static void parse_opts(struct vmur *info, int argc, char *argv[])
 		ERR_EXIT("Missing command\n");
 
 	if (set_action(info, argv[1])) {
-		parse_opts_default(info, argc, argv);
+		parse_opts_default(argc, argv);
 		ERR("Unknown command '%s'\n", argv[1]);
 		std_usage_exit();
 	}
@@ -1067,7 +1312,7 @@ static void check_hold_state(char *spoolid)
 	char cmd[MAXCMDLEN];
 	char *response;
 
-	sprintf(cmd, "QUERY READER * ALL %s", spoolid);
+	sprintf(cmd, "QUERY READER * %s ALL SHORTDATE", spoolid);
 	cpcmd(cmd, &response, 0, 0);
 	response[114] = 0;
 	if (strcmp(&response[110], "NONE") != 0)
@@ -1099,7 +1344,7 @@ static void order_change_reader_file(struct vmur *info)
 /*
  * Issue CP command CLOSE RDR
  */
-static void close_reader(struct vmur *info, char *hold)
+static void close_reader(struct vmur *info, const char *hold)
 {
 	char cmd[MAXCMDLEN];
 
@@ -1107,30 +1352,6 @@ static void close_reader(struct vmur *info, char *hold)
 	cpcmd(cmd, NULL, NULL, 0);
 }
 
-/*
- * strip leading and trailing blanks
- */
-static char *strstrip(char *s)
-{
-	size_t size;
-	char *end;
-
-	size = strlen(s);
-
-	if (!size)
-		return s;
-
-	end = s + size - 1;
-	while (end >= s && isspace(*end))
-		end--;
-	*(end + 1) = '\0';
-
-	while (*s && isspace(*s))
-		s++;
-
-	return s;
-}
-
 /*
  *  Check whether reader can handle class of reader spool file
  */
@@ -1147,7 +1368,7 @@ static int check_class(struct vmur *info)
 	if (device_class == '*') /* reader device can handle any class */
 		return 0;
 
-	sprintf(cmd, "QUERY RDR * %s ALL", info->spoolid);
+	sprintf(cmd, "QUERY RDR * %s ALL SHORTDATE", info->spoolid);
 	cpcmd(cmd, &buf, NULL, 0);
 	file_class = buf[91];
 	free(buf);
@@ -1157,7 +1378,7 @@ static int check_class(struct vmur *info)
 }
 
 /*
- * Get filename for spoolid from reader
+ * Get file name for spoolid from reader
  */
 static int get_filename_from_reader(struct vmur *info)
 {
@@ -1166,12 +1387,12 @@ static int get_filename_from_reader(struct vmur *info)
 	char type[9] = {0};
 	char *buf;
 
-	sprintf(cmd, "QUERY RDR * %s ALL", info->spoolid);
+	sprintf(cmd, "QUERY RDR * %s ALL SHORTDATE", info->spoolid);
 	cpcmd(cmd, &buf, NULL, 0);
 	memcpy(name, &buf[130], 8);
-	strstrip(name);
+	util_strstrip(name);
 	memcpy(type, &buf[140], 8);
-	strstrip(type);
+	util_strstrip(type);
 	free(buf);
 	if (strlen(name) == 0) {
 		ERR("Please specify name to receive "
@@ -1196,16 +1417,15 @@ static int get_filename_from_reader(struct vmur *info)
  */
 static void acquire_lock(struct vmur *info)
 {
-	int fd;
 	char failed_action[10] = {};
 
-	fd = open(LOCK_FILE, O_RDONLY | O_CREAT, S_IRUSR);
-	if (fd == -1) {
+	info->lock_fd = open(LOCK_FILE, O_RDONLY | O_CREAT, S_IRUSR);
+	if (info->lock_fd == -1) {
 		ERR("WARNING: Unable to open lock file %s, continuing "
 		    "without any serialization.\n", LOCK_FILE);
 		return;
 	}
-	if (flock(fd, LOCK_EX | LOCK_NB) == -1) {
+	if (flock(info->lock_fd, LOCK_EX | LOCK_NB) == -1) {
 		switch (info->action) {
 		case RECEIVE:
 			strcpy(failed_action, "received");
@@ -1231,12 +1451,23 @@ static void acquire_lock(struct vmur *info)
 	}
 }
 
+/*
+ * Release lock and close lock file
+ */
+static void release_lock(struct vmur *info)
+{
+	flock(info->lock_fd, LOCK_UN);
+	close(info->lock_fd);
+	info->lock_fd = -1;
+}
+
 /*
  * Check if file already exists. If yes, ask if it should be overwritten.
  */
 static void check_overwrite(struct vmur *info)
 {
 	char buf[5] = {};
+	char *str;
 	struct stat stat_info;
 
 	if (info->force_specified)
@@ -1251,7 +1482,9 @@ static void check_overwrite(struct vmur *info)
 	}
 
 	fprintf(stderr, "%s: Overwrite '%s'? ", prog_name, info->file_name);
-	fgets(buf, sizeof(buf), stdin);
+	str = fgets(buf, sizeof(buf), stdin);
+	if (!str)
+		exit(1);
 	if (strcasecmp(buf, "y\n") == 0)
 		return;
 	if (strcasecmp(buf, "yes\n") == 0)
@@ -1269,7 +1502,7 @@ enum spoolfile_fmt get_spoolfile_fmt(struct vmur *info,
 {
 	struct splink_record *rec;
 	char netdata_id[5] = {0xc9, 0xd5, 0xd4, 0xd9, 0xf0}; /* EBCDIC: INMR0 */
-	int i;
+	unsigned int i;
 
 	rec = (struct splink_record *) &buf->data;
 
@@ -1280,20 +1513,21 @@ enum spoolfile_fmt get_spoolfile_fmt(struct vmur *info,
 
 	for (i = 0; i < buf->data_recs; i++) {
 		if (rec->ccw.flag & CCW_IMMED_FLAG) {
-			rec = ((void *) rec) + sizeof(rec->ccw);
+			rec = (struct splink_record *) ((char *) rec +
+							sizeof(rec->ccw));
 			continue; /* skip immediate CCWs */
 		}
 		if ((rec->data.flag & IS_CONTROL_RECORD) &&
 		    (memcmp(rec->data.magic, netdata_id, 5) == 0))
 			return TYPE_NETDATA;
-		rec = ((void *) rec) + rec->record_len;
+		rec = (struct splink_record *) ((char *) rec + rec->record_len);
 	}
 
 	return TYPE_NORMAL;
 }
 
 /*
- * Convert record for texmode: Do EBCDIC->ASCII translation
+ * Convert record for text mode: Do EBCDIC->ASCII translation
  */
 static int convert_text(struct vmur *info, struct splink_record *rec,
 			char **out_ptr)
@@ -1321,7 +1555,7 @@ out:
 /*
  * Convert record for binary mode: Fill up missing 0x40 bytes
  */
-static int convert_binary(struct vmur *info, struct splink_record *rec,
+static void convert_binary(struct vmur *info, struct splink_record *rec,
 			  char **out_ptr)
 {
 	int residual;
@@ -1335,13 +1569,12 @@ static int convert_binary(struct vmur *info, struct splink_record *rec,
 	residual = info->file_reclen - rec->ccw.data_len;
 	memset(*out_ptr, 0x40, residual);
 	*out_ptr += residual;
-	return 0;
 }
 
 /*
  * Convert record for blocked mode: remove padding bytes and add separator
  */
-static int convert_blocked(struct vmur *info, struct splink_record *rec,
+static void convert_blocked(struct vmur *info, struct splink_record *rec,
 			   char **out_ptr)
 {
 	int residual, i;
@@ -1368,7 +1601,6 @@ static int convert_blocked(struct vmur *info, struct splink_record *rec,
 
 	**out_ptr = info->blocked_separator;
 	*out_ptr += 1;
-	return 0;
 }
 
 /*
@@ -1378,16 +1610,18 @@ ssize_t convert_sfdata(struct vmur *info, struct splink_page *in, char *out)
 {
 	struct splink_record *rec;
 	char *out_ptr = out;
-	int i, rc = 0;
+	unsigned int i, rc = 0;
 
 	rec = (struct splink_record *) &in->data;
 
 	for (i = 0; i < in->data_recs; i++) {
 		if (rec->ccw.opcode == NOP) {
-			rec = ((void *) rec) + rec->record_len;
+			rec = (struct splink_record *) ((char *) rec +
+				rec->record_len);
 			continue; /* skip NOP CCWs */
 		} else if (rec->ccw.flag & CCW_IMMED_FLAG) {
-			rec = ((void *) rec) + sizeof(rec->ccw);
+			rec = (struct splink_record *) ((char *) rec +
+				sizeof(rec->ccw));
 			continue; /* skip immediate CCWs */
 		}
 		if (info->text_specified) {
@@ -1395,15 +1629,10 @@ ssize_t convert_sfdata(struct vmur *info, struct splink_page *in, char *out)
 			if (rc)
 				return rc;
 		} else if (info->blocked_specified) {
-			rc = convert_blocked(info, rec, &out_ptr);
-			if (rc)
-				return rc;
-		} else {
-			rc = convert_binary(info, rec, &out_ptr);
-			if (rc)
-				return rc;
-		}
-		rec = ((void *) rec) + rec->record_len;
+			convert_blocked(info, rec, &out_ptr);
+		} else
+			convert_binary(info, rec, &out_ptr);
+		rec = (struct splink_record *) ((char *) rec + rec->record_len);
 	}
 	return out_ptr - out;
 }
@@ -1421,7 +1650,8 @@ int write_normal(struct vmur *info, struct splink_page *sfdata, int count,
 		struct splink_page *data;
 
 		data = &sfdata[i];
-		outbuf = malloc((info->file_reclen + 1) * data->data_recs);
+		outbuf = (char *) malloc((info->file_reclen + 1) *
+					 data->data_recs);
 		if (!outbuf) {
 			ERR("Out of memory\n");
 			return -ENOMEM;
@@ -1457,12 +1687,23 @@ int write_vmdump(struct vmur *info, struct splink_page *sfdata, int count,
 	return 0;
 }
 
+/*
+ * Clean up and restore spool options
+ */
+static void cleanup_atexit_fn(void)
+{
+	restore_spool_options(&vmur_info);
+	release_lock(&vmur_info);
+}
+
 /*
  * Close VM virtual reader in case of a signal e.g. CTRL-C
 */
-static void ur_receive_sig_handler(int sig, siginfo_t *sip, void *p)
+static void ur_receive_sig_handler(int UNUSED(sig), siginfo_t *UNUSED(sip),
+				   void *UNUSED(p))
 {
 	close_reader(&vmur_info, "HOLD");
+	cleanup_atexit_fn();
 	ERR_EXIT("Operation terminated, spool file received incompletely.\n");
 }
 
@@ -1473,14 +1714,12 @@ static void ur_receive(struct vmur *info)
 {
 	struct splink_page sfdata[READ_BLOCKS];
 	enum spoolfile_fmt type;
-	int fhi, fho, count;
+	int fhi, fho = STDOUT_FILENO, count;
+	int rc;
 
 	if (check_class(info))
-		ERR_EXIT("Reader device class does not match spool file "
-			 "class.\nIssue CP command 'SPOOL %X CLASS *' to "
-			 "enable %X for any spool file class.\n",
-			 info->devno, info->devno);
-	acquire_lock(info);
+		ERR_EXIT("Reader device class does not match the specified "
+			 "spool file class\n");
 	close_reader(info, "HOLD");
 	order_change_reader_file(info);
 
@@ -1491,23 +1730,14 @@ static void ur_receive(struct vmur *info)
 
 	set_signal_handler(info, ur_receive_sig_handler);
 
-	if (info->stdout_specified) {
-		fho = STDOUT_FILENO;
-	} else {
+	if (!info->stdout_specified) {
 		if (!info->file_name_specified &&
 		    get_filename_from_reader(info))
 			goto fail;
 		check_overwrite(info);
-		fho = open(info->file_name, O_WRONLY | O_CREAT | O_TRUNC,
-			   S_IRUSR | S_IWUSR);
-		if (fho == -1) {
-			ERR("Could not open file %s\n%s\n", info->file_name,
-			    strerror(errno));
-			goto fail;
-		}
 	}
 
-	/* Read first block and check spoolfile format */
+	/* Read first block and check spool file format */
 
 	count = read(fhi, &sfdata, READ_BLOCKS * sizeof(sfdata[0]));
 	if (count == -1) {
@@ -1536,8 +1766,17 @@ static void ur_receive(struct vmur *info)
 
 	/* read spool file data, convert it, and write it to output file or
 	 * stdout */
+	if (!info->stdout_specified) {
+		fho = open(info->file_name, O_WRONLY | O_CREAT | O_TRUNC,
+			   S_IRUSR | S_IWUSR);
+		if (fho == -1) {
+			ERR("Could not open file %s\n%s\n", info->file_name,
+			    strerror(errno));
+			goto fail;
+		}
+	}
 	while (count != 0) {
-		int rc, blocks;
+		int blocks;
 
 		blocks = count / sizeof(sfdata[0]);
 		if (type == TYPE_VMDUMP)
@@ -1602,7 +1841,6 @@ static void close_ur_device(struct vmur *info)
 	}
 	cpcmd_cs(cmd, &response, &cprc, 0);
 	memcpy(spoolid, &response[9], 4);
-	free(response);
 	if (cprc == 439) {
 		if (info->action == PUNCH)
 			sprintf(cmd, "PURGE * PUN %s", spoolid);
@@ -1611,7 +1849,10 @@ static void close_ur_device(struct vmur *info)
 		cpcmd(cmd, NULL, NULL, 0);
 		ERR_EXIT("User %s spool fileid limit exceeded.\n"
 			 , info->user);
+	} else if (cprc) {
+		ERR_EXIT("CP command failed with rc=%i\n%s\n", cprc, response);
 	}
+	free(response);
 	if (info->rdr_specified) {
 		if (info->node_specified)
 			printf("Reader file with spoolid %s created and "
@@ -1689,9 +1930,11 @@ static void rscs_punch_setup(struct vmur *info)
 /*
  * Purge punch file in case of a signal e.g. CTRL-C
  */
-static void ur_write_sig_handler(int sig, siginfo_t *sip, void *p)
+static void ur_write_sig_handler(int UNUSED(sig), siginfo_t *UNUSED(sip),
+				 void *UNUSED(p))
 {
 	close_ur_device_purge(&vmur_info);
+	cleanup_atexit_fn();
 	ERR_EXIT("Operation terminated, no spool file created.\n");
 }
 
@@ -1728,15 +1971,16 @@ found:
  */
 static int read_text_file(struct vmur *info, int fd, char *out_buf, size_t len)
 {
-	int rc, pos = 0;
+	unsigned int pos = 0;
 	static int line = 1;
 	char sep, pad;
 	char *buf;
+	int rc;
 
 	sep = '\n';
 	pad = ' ';
 
-	buf = malloc(info->ur_reclen + 1);
+	buf = (char *) malloc(info->ur_reclen + 1);
 	if (!buf)
 		return -ENOMEM;
 
@@ -1781,15 +2025,16 @@ fail:
 static int read_blocked_file(struct vmur *info, int fd, char *out_buf,
 			     size_t len)
 {
-	int line_len, pos = 0;
+	unsigned int pos = 0;
 	static int line = 1;
 	char sep, pad;
+	int line_len;
 	char *buf;
 
 	sep = info->blocked_separator;
 	pad = info->blocked_padding;
 
-	buf = malloc(info->ur_reclen + 1);
+	buf = (char *) malloc(info->ur_reclen + 1);
 	if (!buf)
 		return -ENOMEM;
 
@@ -1801,7 +2046,7 @@ static int read_blocked_file(struct vmur *info, int fd, char *out_buf,
 			ERR("Input line %i too long. Unit record length"
 			    " must not exceed %i\n", line, info->ur_reclen);
 			goto fail;
-		} else if (len < 0) {
+		} else if (line_len < 0) {
 			ERR("Read failed: %s", strerror(errno));
 			goto fail;
 		}
@@ -1848,8 +2093,6 @@ static void ur_write(struct vmur *info)
 	char *sfdata;
 	ssize_t count;
 
-	acquire_lock(info);
-
 	/* close punch preventively */
 	close_ur_device_simple(info);
 
@@ -1858,7 +2101,7 @@ static void ur_write(struct vmur *info)
 		ERR_EXIT("Virtual punch device %X is spooled CONT.\n",
 			 info->devno);
 
-	sfdata = malloc(info->ur_reclen * VMUR_REC_COUNT);
+	sfdata = (char *) malloc(info->ur_reclen * VMUR_REC_COUNT);
 	if (!sfdata)
 		ERR_EXIT("Could allocate memory for buffer (%i)\n",
 			    info->ur_reclen);
@@ -1929,16 +2172,23 @@ fail:
 static void ur_purge_question(struct vmur *info)
 {
 	char buf[5] = {};
+	char *str;
+
 	if (info->force_specified)
 		return;
 
-	if (info->spoolid_specified)
-		fprintf(stderr, "%s: purge %s %s? ", prog_name, info->queue,
-		       info->spoolid);
-	else
-		fprintf(stderr, "%s: purge all %s files? ", prog_name,
-			info->queue);
-	fgets(buf, sizeof(buf), stdin);
+	fprintf(stderr, "%s: purge selected %s file(s)? ", prog_name,
+		info->queue);
+	/*
+	 * Release the vmur session lock while waiting for user input.  It is
+	 * safe to release the lock here, because the ur_purge operation does
+	 * not modify spool options.
+	 */
+	release_lock(info);
+	str = fgets(buf, sizeof(buf), stdin);
+	acquire_lock(info);
+	if (!str)
+		exit(1);
 	if (strcasecmp(buf, "y\n") == 0)
 		return;
 	if (strcasecmp(buf, "yes\n") == 0)
@@ -1952,13 +2202,29 @@ static void ur_purge_question(struct vmur *info)
 static int ur_purge(struct vmur *info)
 {
 	char *buf, cmd[MAXCMDLEN];
+	int n, m;
 
 	ur_purge_question(info);
+
+	/* Prepare the CP PURGE command */
+	n = m = sprintf(cmd, "PURGE * %s", info->queue);
+
+	/* Add selection criteria to match spool files */
 	if (info->spoolid_specified)
-		sprintf(cmd, "PURGE * %s %s", info->queue, info->spoolid);
-	else
-		sprintf(cmd, "PURGE * %s ALL", info->queue);
-	acquire_lock(info);
+		n += sprintf(cmd + n, " %s", info->spoolid);
+	if (info->spool_class_specified)
+		n += sprintf(cmd + n, " CLASS %c", info->spool_class);
+	if (info->spool_form_specified)
+		n += sprintf(cmd + n, " FORM %s", info->spool_form);
+	if (info->spool_dest_specified)
+		n += sprintf(cmd + n, " DEST %s", info->spool_dest);
+
+	/* Purge all files if no spoolid or any of the class, form, or dest
+	 * option is specified.
+	 */
+	if (n == m)
+		n += sprintf(cmd + n, " ALL");
+
 	cpcmd(cmd, &buf, NULL, 0);
 	ERR("%s", buf);
 	free(buf);
@@ -1973,7 +2239,6 @@ static int ur_order(struct vmur *info)
 	char cmd[MAXCMDLEN];
 
 	sprintf(cmd, "ORDER * %s %s", info->queue, info->spoolid);
-	acquire_lock(info);
 	cpcmd(cmd, NULL, NULL, 0);
 	return 0;
 }
@@ -2000,7 +2265,7 @@ static int ur_list(struct vmur *info)
 /*
  * Initialize iconv: "from" -> "to"
  */
-static void setup_iconv(struct vmur *info, char *from, char *to)
+static void setup_iconv(struct vmur *info, const char *from, const char *to)
 {
 	info->iconv = iconv_open(to, from);
 	if (info->iconv == ((iconv_t) -1))
@@ -2010,18 +2275,29 @@ static void setup_iconv(struct vmur *info, char *from, char *to)
 
 int main(int argc, char **argv)
 {
-	/* set name of program */
+	/* Set name of program */
 	prog_name = basename(argv[0]);
 
-	/* set default values */
+	/* Set default values */
 	init_info(&vmur_info);
 
-	/* parse command line options and check syntax */
+	/* Parse command line options and check syntax */
 	parse_opts(&vmur_info, argc, argv);
 
+	/* Register cleanup function */
+	if (atexit(cleanup_atexit_fn))
+		ERR_EXIT("Could not set up vmur session cleanup\n");
+
+	/* Acquire a lock to serialize concurrent vmur invocations */
+	acquire_lock(&vmur_info);
+
+	/* Retrieve ur device number */
+	setup_ur_device(&vmur_info);
+
 	switch (vmur_info.action) {
 	case RECEIVE:
-		setup_ur_device(&vmur_info);
+		/* Setup spool options */
+		setup_spool_options(&vmur_info);
 		if (vmur_info.text_specified)
 			setup_iconv(&vmur_info, EBCDIC_CODE_PAGE,
 				    ASCII_CODE_PAGE);
@@ -2029,10 +2305,11 @@ int main(int argc, char **argv)
 		break;
 	case PUNCH:
 	case PRINT:
+		/* Setup spool options */
+		setup_spool_options(&vmur_info);
 		if (vmur_info.text_specified)
 			setup_iconv(&vmur_info, ASCII_CODE_PAGE,
 				    EBCDIC_CODE_PAGE);
-		setup_ur_device(&vmur_info);
 		ur_write(&vmur_info);
 		break;
 	case PURGE:
