@@ -1,7 +1,7 @@
 /**
  * lszcrypt - Display zcrypt devices and configuration settings
  *
- * Copyright IBM Corp. 2008, 2020
+ * Copyright IBM Corp. 2008, 2022
  *
  * s390-tools is free software; you can redistribute it and/or modify
  * it under the terms of the MIT license. See LICENSE for details.
@@ -29,6 +29,11 @@
  */
 static struct lszcrypt_l {
 	int verbose;
+	int showaccel;
+	int showcca;
+	int showep11;
+	int showcard;
+	int showqueue;
 } l;
 
 /*
@@ -39,15 +44,18 @@ static struct lszcrypt_l {
 #define CAP_CCA		"CCA Secure Key"
 #define CAP_RNG		"Long RNG"
 #define CAP_EP11	"EP11 Secure Key"
+#define CAP_APMMS       "AP bus max message size limit %ld Kb"
+#define CAP_HSL         "Hardware support for stateless filtering"
 
 /*
- * Card types
+ * Card types and other feature masks
  */
 #define MASK_APSC	    0x80000000
 #define MASK_RSA4K	    0x60000000
 #define MASK_COPRO	    0x10000000
 #define MASK_ACCEL	    0x08000000
 #define MASK_EP11	    0x04000000
+#define MASK_HSL            0x01000000
 
 /*
  * Classification
@@ -60,20 +68,21 @@ static struct lszcrypt_l {
 /*
  * facility bits
  */
-#define MAX_FAC_BITS 9
+#define MAX_FAC_BITS 10
 static struct fac_bits_s {
 	int mask;
 	char c;
 } fac_bits[MAX_FAC_BITS] = {
-	{ 0x80000000, 'S' },
-	{ 0x40000000, 'M' },
-	{ 0x20000000, 'C' },
-	{ 0x10000000, 'D' },
-	{ 0x08000000, 'A' },
-	{ 0x04000000, 'X' },
-	{ 0x02000000, 'N' },
-	{ 0x00800000, 'F' },
-	{ 0x00400000, 'R' },
+	{ 0x80000000, 'S' }, /* bit 0 */
+	{ 0x40000000, 'M' }, /* bit 1 */
+	{ 0x20000000, 'C' }, /* bit 2 */
+	{ 0x10000000, 'D' }, /* bit 3, cca mode */
+	{ 0x08000000, 'A' }, /* bit 4, accel mode */
+	{ 0x04000000, 'X' }, /* bit 5, ep11 mode */
+	{ 0x02000000, 'N' }, /* bit 6, apxa */
+	{ 0x01000000, 'H' }, /* bit 7, stateless filtering by hardware */
+	{ 0x00800000, 'F' }, /* bit 8, full function set */
+	{ 0x00400000, 'R' }, /* bit 9, restricted function set */
 };
 
 /*
@@ -95,6 +104,13 @@ static const struct util_prg prg = {
 /*
  * Configuration of command line options
  */
+
+#define OPT_ACCELONLY  0x81
+#define OPT_CCAONLY    0x82
+#define OPT_EP11ONLY   0x83
+#define OPT_CARDONLY   0x84
+#define OPT_QUEUEONLY  0x85
+
 static struct util_opt opt_vec[] = {
 	{
 		.option = {"bus", 0, NULL, 'b'},
@@ -113,6 +129,35 @@ static struct util_opt opt_vec[] = {
 		.option = {"verbose", 0, NULL, 'V'},
 		.desc = "Print verbose messages",
 	},
+	{
+		.option = {"accelonly", 0, NULL, OPT_ACCELONLY},
+		.flags = UTIL_OPT_FLAG_NOSHORT,
+		.desc = "Show only information from cards/queues in Accelerator mode",
+	},
+	{
+		.option = {"ccaonly", 0, NULL, OPT_CCAONLY},
+		.flags = UTIL_OPT_FLAG_NOSHORT,
+		.desc = "Show only information from cards/queues in CCA-Coprocessor mode",
+	},
+	{
+		.option = {"ep11only", 0, NULL, OPT_EP11ONLY},
+		.flags = UTIL_OPT_FLAG_NOSHORT,
+		.desc = "Show only information from cards/queues in EP11-Coprocessor mode",
+	},
+	{
+		.option = {"cardonly", 0, NULL, OPT_CARDONLY},
+		.flags = UTIL_OPT_FLAG_NOSHORT,
+		.desc = "Show only information from cards but no queue info",
+	},
+	{
+		.option = {"queueonly", 0, NULL, OPT_QUEUEONLY},
+		.flags = UTIL_OPT_FLAG_NOSHORT,
+		.desc = "Show only information from queues but no card info",
+	},
+	{
+		.option = {"serial", 0, NULL, 's'},
+		.desc = "Show the serial numbers for CCA and EP11 crypto cards",
+	},
 	UTIL_OPT_HELP,
 	UTIL_OPT_VERSION,
 	UTIL_OPT_END
@@ -247,13 +292,94 @@ static void show_domains(void)
 	show_domains_util_rec(domain_array);
 }
 
+/*
+ * Show serialnumbers
+ */
+static void show_serialnumbers(void)
+{
+	struct util_rec *rec = util_rec_new_wide("-");
+	struct dirent **dev_vec;
+	int i, count;
+	char *ap, *path, *device, *grp_dev, card[16], buf[256];
+	long config = -1, online = -1, chkstop = -1;
+	unsigned long facility;
+
+	/* check if ap driver is available */
+	ap = util_path_sysfs("bus/ap");
+	if (!util_path_is_dir(ap))
+		errx(EXIT_FAILURE, "Crypto device driver not available.");
+
+	/* define the record */
+	util_rec_def(rec, "card", UTIL_REC_ALIGN_LEFT, 8, "CARD.DOM");
+	util_rec_def(rec, "type", UTIL_REC_ALIGN_LEFT, 5, "TYPE");
+	util_rec_def(rec, "mode", UTIL_REC_ALIGN_LEFT, 11, "MODE");
+	util_rec_def(rec, "status", UTIL_REC_ALIGN_LEFT, 10, "STATUS");
+	util_rec_def(rec, "serialnr", UTIL_REC_ALIGN_LEFT, 8, "SERIALNR");
+
+	/* scan the devices */
+	path = util_path_sysfs("devices/ap/");
+	count = util_scandir(&dev_vec, alphasort, path, "card[0-9a-fA-F]+");
+	if (count < 1)
+		errx(EXIT_FAILURE, "No crypto card devices found.");
+	util_rec_print_hdr(rec);
+	for (i = 0; i < count; i++) {
+		device = dev_vec[i]->d_name;
+		grp_dev = util_path_sysfs("devices/ap/%s", device);
+		if (!util_path_is_dir(grp_dev))
+			errx(EXIT_FAILURE, "Error - cryptographic device %s does not exist.", device);
+		if (!util_path_is_readable("%s/type", grp_dev) ||
+		    !util_path_is_readable("%s/online", grp_dev))
+			goto next;
+		strcpy(card, device + 4);
+		util_rec_set(rec, "card", card);
+		util_file_read_line(buf, sizeof(buf), "%s/type", grp_dev);
+		util_rec_set(rec, "type", buf);
+		util_file_read_ul(&facility, 16, "%s/ap_functions", grp_dev);
+		if (facility & MASK_COPRO)
+			util_rec_set(rec, "mode", "CCA-Coproc");
+		else if (facility & MASK_EP11)
+			util_rec_set(rec, "mode", "EP11-Coproc");
+		else
+			goto next;
+		if (util_path_is_readable("%s/config", grp_dev))
+			util_file_read_l(&config, 10, "%s/config", grp_dev);
+		if (util_path_is_readable("%s/chkstop", grp_dev))
+			util_file_read_l(&chkstop, 10, "%s/chkstop", grp_dev);
+		if (util_path_is_readable("%s/online", grp_dev))
+			util_file_read_l(&online, 10, "%s/online", grp_dev);
+		if (config == 0) {
+			util_rec_set(rec, "status", "deconfig");
+		} else {
+			if (chkstop > 0)
+				util_rec_set(rec, "status", "chkstop");
+			else if (online > 0)
+				util_rec_set(rec, "status", "online");
+			else if (online == 0)
+				util_rec_set(rec, "status", "offline");
+			else
+				util_rec_set(rec, "status", "-");
+		}
+		if (util_file_read_line(buf, sizeof(buf), "%s/serialnr", grp_dev))
+			util_rec_set(rec, "serialnr", "-");
+		else {
+			buf[8] = '\0';
+			util_rec_set(rec, "serialnr", buf);
+		}
+		util_rec_print(rec);
+next:
+		free(grp_dev);
+	}
+
+	free(path);
+}
+
 /*
  * Show capability
  */
 static void show_capability(const char *id_str)
 {
 	unsigned long func_val;
-	long hwtype, id;
+	long hwtype, id, max_msg_size;
 	char *p, *ap, *dev, card[16], cbuf[256];
 
 	/* check if ap driver is available */
@@ -272,6 +398,9 @@ static void show_capability(const char *id_str)
 	/* If sysfs attribute is missing, set functions to 0 */
 	if (util_file_read_ul(&func_val, 16, "%s/ap_functions", dev))
 		func_val = 0x00000000;
+	/* try to read the ap bus max message size for this card */
+	if (util_file_read_l(&max_msg_size, 10, "%s/max_msg_size", dev))
+		max_msg_size = 0;
 	/* Skip devices, which are not supported by zcrypt layer */
 	if (!util_path_is_readable("%s/type", dev) ||
 	    !util_path_is_readable("%s/online", dev)) {
@@ -289,9 +418,9 @@ static void show_capability(const char *id_str)
 	case 6:
 	case 8:
 		if (func_val & MASK_RSA4K)
-			printf("%s", CAP_RSA4K);
+			printf("%s\n", CAP_RSA4K);
 		else
-			printf("%s", CAP_RSA2K);
+			printf("%s\n", CAP_RSA2K);
 		break;
 	case 7:
 	case 9:
@@ -300,37 +429,41 @@ static void show_capability(const char *id_str)
 			printf("%s (%s)\n", CAP_CCA, cbuf);
 		else
 			printf("%s\n", CAP_CCA);
-		printf("%s", CAP_RNG);
+		printf("%s\n", CAP_RNG);
 		break;
 	case 10: /* CEX4S */
 	case 11: /* CEX5S */
 	case 12: /* CEX6S */
 	case 13: /* CEX7S */
+	case 14: /* CEX8S */
 		if (func_val & MASK_ACCEL) {
 			if (func_val & MASK_RSA4K)
-				printf("%s", CAP_RSA4K);
+				printf("%s\n", CAP_RSA4K);
 			else
-				printf("%s", CAP_RSA2K);
+				printf("%s\n", CAP_RSA2K);
 		} else if (func_val & MASK_COPRO) {
 			printf("%s\n", CAP_RSA4K);
 			if (cbuf[0])
 				printf("%s (%s)\n", CAP_CCA, cbuf);
 			else
 				printf("%s\n", CAP_CCA);
-			printf("%s", CAP_RNG);
+			if (func_val & MASK_HSL)
+				printf("%s\n", CAP_HSL);
+			printf("%s\n", CAP_RNG);
 		} else if (func_val & MASK_EP11) {
-			printf("%s", CAP_EP11);
+			printf("%s\n", CAP_EP11);
 		} else {
-			printf("Detailed capability information for %s (hardware type %ld) is not available.",
+			printf("Detailed capability information for %s (hardware type %ld) is not available.\n",
 			       card, hwtype);
 		}
+		if (max_msg_size > 0)
+			printf(CAP_APMMS "\n", max_msg_size / 1024);
 		break;
 	default:
-			printf("Detailed capability information for %s (hardware type %ld) is not available.",
-			       card, hwtype);
+		printf("Detailed capability information for %s (hardware type %ld) is not available.\n",
+		       card, hwtype);
 		break;
 	}
-	printf("\n");
 }
 
 /*
@@ -375,7 +508,7 @@ static int read_driver(const char *dir, const char *subdir, char *buf, size_t bu
 static void read_subdev_rec_default(struct util_rec *rec, const char *grp_dev,
 				    const char *sub_dev)
 {
-	long config = -1, online = -1;
+	long config = -1, online = -1, chkstop = -1;
 	char buf[256];
 	unsigned long facility;
 
@@ -386,6 +519,8 @@ static void read_subdev_rec_default(struct util_rec *rec, const char *grp_dev,
 
 	if (util_path_is_readable("%s/%s/config", grp_dev, sub_dev))
 		util_file_read_l(&config, 10, "%s/%s/config", grp_dev, sub_dev);
+	if (util_path_is_readable("%s/%s/chkstop", grp_dev, sub_dev))
+		util_file_read_l(&chkstop, 10, "%s/%s/chkstop", grp_dev, sub_dev);
 	if (util_path_is_readable("%s/%s/online", grp_dev, sub_dev))
 		util_file_read_l(&online, 10, "%s/%s/online", grp_dev, sub_dev);
 
@@ -393,7 +528,9 @@ static void read_subdev_rec_default(struct util_rec *rec, const char *grp_dev,
 	if (config == 0) {
 		util_rec_set(rec, "status", "deconfig");
 	} else {
-		if (online > 0)
+		if (chkstop > 0)
+			util_rec_set(rec, "status", "chkstop");
+		else if (online > 0)
 			util_rec_set(rec, "status", "online");
 		else if (online == 0)
 			util_rec_set(rec, "status", "offline");
@@ -489,6 +626,9 @@ static void show_subdevice(struct util_rec *rec, const char *grp_dev,
 	     !util_path_is_readable("%s/%s/online", grp_dev, sub_dev)))
 		return;
 
+	if (!l.showqueue)
+		return;
+
 	util_rec_set(rec, "card", sub_dev);
 	read_subdev_rec_default(rec, grp_dev, sub_dev);
 	read_subdev_rec_verbose(rec, grp_dev, sub_dev);
@@ -516,7 +656,7 @@ static void show_subdevices(struct util_rec *rec, const char *grp_dev)
  */
 static void read_rec_default(struct util_rec *rec, const char *grp_dev)
 {
-	long config = -1, online = -1;
+	long config = -1, online = -1, chkstop = -1;
 	char buf[256];
 	unsigned long facility;
 
@@ -537,12 +677,16 @@ static void read_rec_default(struct util_rec *rec, const char *grp_dev)
 
 	if (util_path_is_readable("%s/config", grp_dev))
 		util_file_read_l(&config, 10, "%s/config", grp_dev);
+	if (util_path_is_readable("%s/chkstop", grp_dev))
+		util_file_read_l(&chkstop, 10, "%s/chkstop", grp_dev);
 	if (util_path_is_readable("%s/online", grp_dev))
 		util_file_read_l(&online, 10, "%s/online", grp_dev);
 	if (config == 0) {
 		util_rec_set(rec, "status", "deconfig");
 	} else {
-		if (online > 0)
+		if (chkstop > 0)
+			util_rec_set(rec, "status", "chkstop");
+		else if (online > 0)
 			util_rec_set(rec, "status", "online");
 		else if (online == 0)
 			util_rec_set(rec, "status", "offline");
@@ -603,7 +747,7 @@ static void read_rec_verbose(struct util_rec *rec, const char *grp_dev)
  */
 static void show_device(struct util_rec *rec, const char *device)
 {
-	char *grp_dev, card[16];
+	char *grp_dev, card[16], type[16], t = '\0';
 
 	strcpy(card, &device[4]);
 	grp_dev = util_path_sysfs("devices/ap/%s", device);
@@ -621,11 +765,21 @@ static void show_device(struct util_rec *rec, const char *device)
 	}
 	util_rec_set(rec, "card", card);
 
+	if (util_file_read_line(type, sizeof(type), "%s/type", grp_dev) == 0)
+		t = type[strlen(type) - 1];
+
+	if ((t == 'A' && !l.showaccel) ||
+	    (t == 'C' && !l.showcca) ||
+	    (t == 'P' && !l.showep11))
+		goto out_free;
+
 	read_rec_default(rec, grp_dev);
 	read_rec_verbose(rec, grp_dev);
 
-	util_rec_print(rec);
-	show_subdevices(rec, grp_dev);
+	if (l.showcard)
+		util_rec_print(rec);
+	if (l.showqueue)
+		show_subdevices(rec, grp_dev);
 out_free:
 	free(grp_dev);
 }
@@ -723,7 +877,7 @@ static void show_devices_argv(char *argv[])
 				show_device(rec, card);
 				free(card);
 			}
-			return;
+			continue;
 		}
 		if (sscanf(argv[i]+1, "%x", &dom) == 1) {
 			/* list specific domains of all adapters */
@@ -748,7 +902,7 @@ static void show_devices_argv(char *argv[])
 				}
 				free(path);
 			}
-			return;
+			continue;
 		}
 		printf("Invalid adpater id!\n");
 	}
@@ -799,9 +953,27 @@ int main(int argc, char **argv)
 		case 'd':
 			show_domains();
 			return EXIT_SUCCESS;
+		case 's':
+			show_serialnumbers();
+			return EXIT_SUCCESS;
 		case 'V':
 			l.verbose++;
 			break;
+		case OPT_ACCELONLY:
+			l.showaccel = 1;
+			break;
+		case OPT_CCAONLY:
+			l.showcca = 1;
+			break;
+		case OPT_EP11ONLY:
+			l.showep11 = 1;
+			break;
+		case OPT_CARDONLY:
+			l.showcard = 1;
+			break;
+		case OPT_QUEUEONLY:
+			l.showqueue = 1;
+			break;
 		case 'h':
 			util_prg_print_help();
 			util_opt_print_help();
@@ -815,6 +987,29 @@ int main(int argc, char **argv)
 			return EXIT_FAILURE;
 		}
 	}
+
+	switch (l.showaccel + l.showcca + l.showep11) {
+	case 0:
+		l.showaccel = l.showcca = l.showep11 = 1;
+		break;
+	case 1:
+		break;
+	default:
+		warnx("Only one of --accelonly or --ccaonly or --ep11only can be specified");
+		return EXIT_FAILURE;
+	}
+
+	switch (l.showcard + l.showqueue) {
+	case 0:
+		l.showcard = l.showqueue = 1;
+		break;
+	case 1:
+		break;
+	default:
+		warnx("Only one of --cardonly or --queueonly can be specified");
+		return EXIT_FAILURE;
+	}
+
 	if (optind == argc)
 		show_devices_all();
 	else
