@@ -3,7 +3,7 @@
  *
  * Generic input dump format functions (DFI - Dump Format Input)
  *
- * Copyright IBM Corp. 2001, 2018
+ * Copyright IBM Corp. 2001, 2023
  *
  * s390-tools is free software; you can redistribute it and/or modify
  * it under the terms of the MIT license. See LICENSE for details.
@@ -26,6 +26,7 @@
  * DFI vector - ensure that tape is the first in the list and devmem the second!
  */
 static struct dfi *dfi_vec[] = {
+	/* clang-format off */
 	&dfi_s390tape,
 	&dfi_devmem,
 	&dfi_s390mv_ext,
@@ -33,11 +34,14 @@ static struct dfi *dfi_vec[] = {
 	&dfi_s390_ext,
 	&dfi_s390,
 	&dfi_lkcd,
+	&dfi_vmdump,
+	&dfi_pv_elf,
 	&dfi_elf,
 	&dfi_kdump,
 	&dfi_kdump_flat,
 	&dfi_ngdump,
 	NULL,
+	/* clang-format on */
 };
 
 /*
@@ -168,6 +172,11 @@ struct dfi_cpu *dfi_cpu_alloc(void)
 	return zg_alloc(sizeof(struct dfi_cpu));
 }
 
+void dfi_cpu_free(struct dfi_cpu *cpu)
+{
+	zg_free(cpu);
+}
+
 /*
  * Add DFI CPU
  */
@@ -780,7 +789,7 @@ int dfi_init(void)
 		}
 		util_log_print(UTIL_LOG_DEBUG, "DFI %s returned with rc %d\n",
 			       dfi->name, rc);
-		if (rc == 0 || rc == -EINVAL)
+		if (rc == 0 || rc == -EINVAL || rc == -ENOKEY)
 			return rc;
 		zg_close(g.fh);
 		i++;
