@@ -10,6 +10,7 @@
  *
  */
 
+static const char *VERSION_KEYWORD = "version";
 
 /* Need ISOC99 function isblank() in ctype.h */
 #ifndef __USE_ISOC99
@@ -40,52 +41,51 @@
 #include "misc.h"
 #include "scan.h"
 
-
 /* Determines which keyword may be present in which section */
 enum scan_key_state scan_key_table[SCAN_SECTION_NUM][SCAN_KEYWORD_NUM] = {
 /*	 defa dump dump imag para parm ramd segm targ prom time defa tape mv
  *	 ult  to   tofs e    mete file isk  ent  et   pt   out  ultm      dump
  *			     rs                                 enu
  *
- *       targ targ targ targ targ defa kdum secu
- *       etba etty etge etbl etof ulta p    re
+ *       targ targ targ targ targ defa kdum secu opti
+ *       etba etty etge etbl etof ulta p    re   onal
  *       se   pe   omet ocks fset uto
  *                 ry   ize
  */
 /* default auto */
 	{opt, inv, inv, inv, inv, inv, inv, inv, req, opt, opt, inv, inv, inv,
-	 opt, opt, opt, opt, opt, opt, inv, opt},
+	 opt, opt, opt, opt, opt, opt, inv, opt, inv},
 /* default menu */
 	{inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, req, inv, inv,
-	 inv, inv, inv, inv, inv, inv, inv, opt},
+	 inv, inv, inv, inv, inv, inv, inv, opt, inv},
 /* default section */
 	{req, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv,
-	 inv, inv, inv, inv, inv, inv, inv, opt},
+	 inv, inv, inv, inv, inv, inv, inv, opt, inv},
 /* ipl		*/
 	{inv, inv, inv, req, opt, opt, opt, inv, req, inv, inv, inv, inv, inv,
-	 opt, opt, opt, opt, opt, inv, opt, opt},
+	 opt, opt, opt, opt, opt, inv, opt, opt, opt},
 /* segment load */
 	{inv, inv, inv, inv, inv, inv, inv, req, req, inv, inv, inv, inv, inv,
-	 inv, inv, inv, inv, inv, inv, inv, inv},
+	 inv, inv, inv, inv, inv, inv, inv, inv, inv},
 /* part dump	*/
 	{inv, req, inv, inv, inv, inv, inv, inv, opt, inv, inv, inv, inv, inv,
-	 inv, inv, inv, inv, inv, inv, inv, inv},
+	 inv, inv, inv, inv, inv, inv, inv, inv, inv},
 /* fs dump	*/
 	{inv, inv, req, inv, opt, opt, inv, inv, req, inv, inv, inv, inv, inv,
-	 inv, inv, inv, inv, inv, inv, inv, inv},
+	 inv, inv, inv, inv, inv, inv, inv, inv, inv},
 /* ipl tape	*/
 	{inv, inv, inv, req, opt, opt, opt, inv, inv, inv, inv, inv, req, inv,
-	 inv, inv, inv, inv, inv, inv, inv, inv},
+	 inv, inv, inv, inv, inv, inv, inv, inv, inv},
 /* multi volume dump	*/
 	{inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, inv, req,
-	 inv, inv, inv, inv, inv, inv, inv, inv}
+	 inv, inv, inv, inv, inv, inv, inv, inv, inv}
 };
 
 /* Determines which keyword may be present in a menu section */
 static enum scan_key_state scan_menu_key_table[SCAN_KEYWORD_NUM] = {
 /* menu section */
 	 opt, inv, inv, inv, inv, inv, inv, inv, req, opt, opt, inv, inv, inv,
-	 opt, opt, opt, opt, opt, inv, inv, opt
+	 opt, opt, opt, opt, opt, inv, inv, opt, inv
 };
 
 /* Mapping of keyword IDs to strings */
@@ -114,6 +114,7 @@ static const struct {
 	{ "tape", scan_keyword_tape},
 	{ "kdump", scan_keyword_kdump},
 	{ "secure", scan_keyword_secure},
+	{ "optional", scan_keyword_optional},
 };
 
 /* List of keywords that are used without an assignment */
@@ -645,7 +646,7 @@ scan_file(const char* filename, struct scan_token** token)
 
 
 static int
-bls_filter(const struct dirent *ent)
+bls_filter_by_names(const struct dirent *ent)
 {
 	int offset = strlen(ent->d_name) - strlen(".conf");
 
@@ -655,13 +656,111 @@ bls_filter(const struct dirent *ent)
 	return strncmp(ent->d_name + offset, ".conf", strlen(".conf")) == 0;
 }
 
+struct version {
+	char *line; /* pointer to a line with version keyword */
+	int offset; /* offset of version value in the line */
+};
+
+/*
+ * Locate version in bls file represented by ENT
+ */
+static void get_version(const struct dirent *ent, struct version *v)
+{
+	char *line = NULL;
+	size_t len = 0;
+	char *d_name;
+	FILE *stream;
+	ssize_t read;
+
+	memset(v, 0, sizeof(*v));
+	d_name = misc_make_path((char *)blsdir, (char *)ent->d_name);
+	if (!d_name)
+		return;
+
+	stream = fopen(d_name, "r");
+	free(d_name);
+	if (!stream)
+		return;
+
+	while ((read = getline(&line, &len, stream)) != -1) {
+		if (line[read - 1] == '\n') {
+			line[read - 1] = '\0';
+			read--;
+		}
+		if ((size_t)read <= strlen(VERSION_KEYWORD) + 1)
+			continue;
+		if (strcmp(VERSION_KEYWORD, line) > 0)
+			continue;
+		if (!isblank(line[strlen(VERSION_KEYWORD)]))
+			continue;
+		/* skip blanks */
+		v->offset = strlen(VERSION_KEYWORD) + 1;
+		while (v->offset < read - 1 && isblank(line[v->offset]))
+			v->offset++;
+		if (isblank(line[v->offset]))
+			/*
+			 * all characters after the keyword
+			 * are blanks. Invalid version
+			 */
+			continue;
+		v->line = line;
+		fclose(stream);
+		return;
+	}
+	free(line);
+	fclose(stream);
+}
+
+static void put_version(struct version *v)
+{
+	free(v->line);
+}
+
+/**
+ * Check version in bls file represented by ENT.
+ * Return 1 if version is valid. Otherwise return 0
+ */
+static int bls_filter_by_versions(const struct dirent *ent)
+{
+	struct version v;
+
+	if (bls_filter_by_names(ent) == 0)
+		return 0;
+
+	get_version(ent, &v);
+	if (v.line) {
+		put_version(&v);
+		return 1;
+	}
+	return 0;
+}
+
 
 static int
-bls_sort(const struct dirent **ent_a, const struct dirent **ent_b)
+bls_sort_by_names(const struct dirent **ent_a, const struct dirent **ent_b)
 {
 	return strverscmp((*ent_a)->d_name, (*ent_b)->d_name);
 }
 
+static int
+bls_sort_by_versions(const struct dirent **ent_a, const struct dirent **ent_b)
+{
+	struct version v1, v2;
+	int ret;
+
+	get_version(*ent_a, &v1);
+	get_version(*ent_b, &v2);
+	/*
+	 * Both versions are valid.
+	 * It is guaranteed by bls_filter_by_versions()
+	 */
+	ret = strverscmp(v1.line + v1.offset, v2.line + v2.offset);
+
+	put_version(&v1);
+	put_version(&v2);
+
+	return ret;
+}
 
 static int
 scan_append_section_heading(struct scan_token* scan, int* index, char* name);
@@ -794,102 +893,6 @@ static int sort_bls_fields(struct misc_file_buffer *file, char *filename)
 	return 0;
 }
 
-int
-scan_bls(const char* blsdir, struct scan_token** token, int scan_size)
-{
-	int count = 0;
-	int size, remaining = 0, n, current, rc = -1;
-	struct scan_token* buffer;
-	struct scan_token* array = *token;
-	struct dirent** bls_entries;
-	struct misc_file_buffer file;
-	struct stat sb;
-	char filename[PATH_MAX];
-
-	if (!(stat(blsdir, &sb) == 0 && S_ISDIR(sb.st_mode)))
-		return 0;
-
-	n = scandir(blsdir, &bls_entries, bls_filter, bls_sort);
-	if (n <= 0)
-		return n;
-
-	while (array[count].id != 0)
-		count++;
-
-	remaining = scan_size - count;
-
-	/* The array of scanned tokens is allocated when the zipl config file is
-	 * parsed. Its size is a multiple of INITIAL_ARRAY_LENGTH so it may have
-	 * enough space to scan all the tokens that are defined in the BLS files.
-	 * Calculate if is enough assuming that a BLS fragment can contain up to
-	 * 4 tokens: a section heading and 3 keywords (image, ramdisk, parameter).
-	 */
-	if (remaining < n * 4) {
-		size = scan_size - remaining + (n * 4);
-		buffer = (struct scan_token *)misc_malloc(size * sizeof(struct scan_token));
-		if (!buffer)
-			goto err;
-		memset(buffer, 0, size * sizeof(struct scan_token));
-		memcpy(buffer, array, count * sizeof(struct scan_token));
-	} else {
-		buffer = array;
-	}
-
-	while (n--) {
-		sprintf(filename, "%s/%s", blsdir, bls_entries[n]->d_name);
-		printf("Using BLS config file '%s'\n", filename);
-
-		rc = misc_get_file_buffer(filename, &file);
-		if (rc)
-			goto err;
-
-		rc = sort_bls_fields(&file, filename);
-		if (rc)
-			goto err;
-
-		while ((size_t)file.pos < file.length) {
-			current = misc_get_char(&file, 0);
-			switch (current) {
-			case '#':
-				file.pos++;
-				skip_line(&file);
-				break;
-			case EOF:
-				break;
-			case '\t':
-			case '\n':
-			case '\0':
-			case ' ':
-				file.pos++;
-				break;
-			default:
-				rc = scan_bls_field(&file, buffer, &count);
-				if (rc) {
-					error_reason("Incorrect BLS field in "
-						"config file %s\n", filename);
-					goto err;
-				}
-				break;
-			}
-		}
-
-		misc_free_file_buffer(&file);
-		free(bls_entries[n]);
-	}
-
-	*token = buffer;
-	rc = 0;
-err:
-	if (n > 0) {
-		do {
-			free(bls_entries[n]);
-		} while (n-- > 0);
-	}
-
-	free(bls_entries);
-	return rc;
-}
-
 
 /* Search scanned tokens SCAN for a section/menu heading (according to
  * TYPE) of the given NAME, beginning at token OFFSET. Return the index of
@@ -1107,6 +1110,40 @@ scan_count_target_keywords(char* keyword[])
 	return num;
 }
 
+static int bls_scandir(struct dirent ***bls_entries)
+{
+	struct dirent **entries1;
+	struct dirent **entries2;
+	int n1, n2;
+
+	/* arrange by names */
+	n1 = scandir(blsdir, &entries1,
+		     bls_filter_by_names, bls_sort_by_names);
+	if (n1 <= 0)
+		return n1;
+	/* arrange by versions */
+	n2 = scandir(blsdir, &entries2,
+		     bls_filter_by_versions, bls_sort_by_versions);
+
+	if (n2 <= 0 || n2 < n1) {
+		/*
+		 * failed to sort by versions,
+		 * fall back to sorting by filenames
+		 */
+		*bls_entries = entries1;
+		while (n2--)
+			free(entries2[n2]);
+		free(entries2);
+		return n1;
+	}
+	/* use arrangement by versions */
+	*bls_entries = entries2;
+	while (n1--)
+		free(entries1[n1]);
+	free(entries1);
+	return n2;
+}
+
 int
 scan_check_target_data(char* keyword[], int* line)
 {
@@ -1358,6 +1395,163 @@ scan_get_section_keywords(struct scan_token* scan, int* index, char* name,
 }
 
 
+/*
+ * Get the default target defined either in the defaultboot section or the
+ * menu section for the defaultmenu if there is a default menu defined.
+ */
+static char *scan_get_default_target(struct scan_token *scan)
+{
+	int keyword_line[SCAN_KEYWORD_NUM];
+	char *keyword[SCAN_KEYWORD_NUM];
+	int num_line[BOOT_MENU_ENTRIES];
+	char *num[BOOT_MENU_ENTRIES];
+	int i, j, rc;
+
+	/* Find the defaultboot section */
+	i = scan_find_section(scan, DEFAULTBOOT_SECTION,
+			      scan_id_section_heading, 0);
+	if (i < 0) {
+		error_reason("No '%s' section found and no section specified on command line",
+			     DEFAULTBOOT_SECTION);
+		return NULL;
+	}
+
+	/* Get keyword and number data */
+	rc = scan_get_section_keywords(scan, &i, DEFAULTBOOT_SECTION, keyword,
+				       keyword_line, NULL, NULL);
+	if (rc) {
+		error_reason("Could not get keywords for '%s' section",
+			     DEFAULTBOOT_SECTION);
+		return NULL;
+	}
+
+	/* Check if a defaultmenu is set */
+	i = (int) scan_keyword_defaultmenu;
+	if (keyword[i]) {
+		j = scan_find_section(scan, keyword[i],
+				      scan_id_menu_heading, 0);
+		if (j < 0) {
+			error_reason("No '%s' section found", keyword[i]);
+			return NULL;
+		}
+
+		/* Get keyword and number data */
+		rc = scan_get_section_keywords(scan, &j, keyword[i], keyword,
+					       keyword_line, num, num_line);
+		if (rc) {
+			error_reason("Could not get keywords for '%s' section",
+				     keyword[i]);
+			return NULL;
+		}
+	}
+
+	return keyword[(int) scan_keyword_target];
+}
+
+#define BLS_TOKEN_MAX 5 /* section heading, image, ramdisk, parameter, target */
+
+int scan_bls(struct scan_token **token, int scan_size)
+{
+	int size, remaining = 0, n, current, rc = -1, count = 0;
+	struct scan_token *array = *token;
+	struct misc_file_buffer file;
+	struct dirent **bls_entries;
+	struct scan_token *buffer;
+	char filename[PATH_MAX];
+	char *target = NULL;
+	struct stat sb;
+
+	if (!(stat(blsdir, &sb) == 0 && S_ISDIR(sb.st_mode)))
+		return 0;
+
+	n = bls_scandir(&bls_entries);
+	if (n <= 0)
+		return n;
+
+	while (array[count].id != 0)
+		count++;
+
+	remaining = scan_size - count;
+
+	/* The array of scanned tokens is allocated when the zipl config file is
+	 * parsed. Its size is a multiple of INITIAL_ARRAY_LENGTH so it may have
+	 * enough space to scan all the tokens that are defined in the BLS files.
+	 * Calculate if is enough assuming that a BLS fragment can contain up to
+	 * BLS_TOKEN_MAX.
+	 */
+	if (remaining < n * BLS_TOKEN_MAX) {
+		size = scan_size - remaining + (n * BLS_TOKEN_MAX);
+		buffer = (struct scan_token *)misc_malloc(size * sizeof(struct scan_token));
+		if (!buffer)
+			goto err;
+		memset(buffer, 0, size * sizeof(struct scan_token));
+		memcpy(buffer, array, count * sizeof(struct scan_token));
+	} else {
+		buffer = array;
+	}
+
+	target = scan_get_default_target(array);
+	while (n--) {
+		sprintf(filename, "%s/%s", blsdir, bls_entries[n]->d_name);
+		printf("Using BLS config file '%s'\n", filename);
+
+		rc = misc_get_file_buffer(filename, &file);
+		if (rc)
+			goto err;
+
+		rc = sort_bls_fields(&file, filename);
+		if (rc)
+			goto err;
+
+		while ((size_t)file.pos < file.length) {
+			current = misc_get_char(&file, 0);
+			switch (current) {
+			case '#':
+				file.pos++;
+				skip_line(&file);
+				break;
+			case EOF:
+				break;
+			case '\t':
+			case '\n':
+			case '\0':
+			case ' ':
+				file.pos++;
+				break;
+			default:
+				rc = scan_bls_field(&file, buffer, &count);
+				if (rc) {
+					error_reason("Incorrect BLS field in config file %s\n",
+						     filename);
+					goto err;
+				}
+				break;
+			}
+		}
+
+		if (target != NULL)
+			scan_append_keyword_assignment(buffer, &count,
+						       scan_keyword_target,
+						       target);
+
+		misc_free_file_buffer(&file);
+		free(bls_entries[n]);
+	}
+
+	*token = buffer;
+	rc = 0;
+err:
+	if (n > 0) {
+		do {
+			free(bls_entries[n]);
+		} while (n-- > 0);
+	}
+
+	free(bls_entries);
+	return rc;
+}
+
+
 /* Check section at INDEX for compliance with config file rules. Upon success,
  * return zero and advance INDEX to point to the end of the section. Return
  * non-zero otherwise. */
